// ./zcc -lpcre -x xlambda -x xalef-iter -x xaif -x xpcre xoc/test/combined.c

// Interesting bug: If you define multiple functions with the same
// name, it doesn't complain and only outputs the first one.

// Things that can't be combined:
// * Closures can't access $n from the enclosing environment.
//   Currently this actually causes xpcre to panic because
//   xcompilefnsym doesn't get called on the closure function, so
//   lastmatch never gets set, but more fundamentally, these don't
//   look or behave like regular variables, so xlambda can't push them
//   across in the environment.
//
//   For example, replacing the printf with:
//     newfn char* get0(void) {
//       return $0.str;
//     };
//     printf("%s\n", get0());

extern void *memset(void *s, int c, int n);
extern int strlen(const char *s);

extern char *gets(char *s);
extern int strcmp(const char *s1, const char *s2);
extern int printf(const char *format, ...);

#include <pcre.h>
#include "lambda.h"

// Don't show this function
char*
getline()
{
	static char buf[100];
	return gets(buf);
}

void
foreach(char **str, int nstr, void(*f)(char*))
{
    f(str[0::nstr]);
}

int
main(int argc, char **argv)
{
    while(getline()) {
        foreach(argv+1, argc-1,
            fn void check(char *pat) {
                if(it =~ pat)
                    printf("%s\n", $0.str);
            });
        free(check);
    }
}
