// Compile AST into byte codes.

// TODO: 
//
//	classes?
//	garbage collector

// Long term TODO:
//	iterators
//	polymorphism

#include "a.h"
#include "code.h"

DEFINE_bool(tailcalls, true, "Implement tail calls as jumps.");
DEFINE_bool(disassemble, false, "Disassemble generated code.");

Loc *evalinto(Compile*, Expr*, Loc*);
Loc *evaltmp(Compile*, Expr*);
Loc *levalexpr(Compile*, Expr*, int);
Loc *derefmember(Compile*, Line, Loc*, Type*, Expr*, int);
void emitCLL(Compile *cs, Line, Code *code, Loc *l1, Loc *l2);
Loc *cvtbool(Compile*, Line, Loc*);
Loc *simpleassign(Compile*, Expr*);
void assign(Compile *cs, Line line, Expr *left, Loc *right);
Loc* maybelevalexpr(Compile *cs, Expr *expr, int write);

// Allocate a new location structure with the given mode and type
// in cs->arena_. 
Loc*
Compile::localloc(int mode, Type *type)
{
	Loc *l;
	
	l = (Loc*)amalloc(&arena_, sizeof(Loc));
	l->mode = mode;
	l->type = type;
	return l;
}

// Allocate a new slot of the given type on the temporary stack.
Loc*
Compile::mktmp(Type *type)
{
	Loc *l;
	
	if(tmpstack_ == nil)
		panic("mktmp");

	l = localloc(LocFrame, type);
	if(tc_->frame_ == globalframe)
		l->mode = LocGlobal;
	else
		l->mode = LocFrame;
	l->off = tc_->frame_->tmpoff;
	tc_->frame_->tmpoff += sizeoftype(type);
	if(tc_->frame_->tmpoff > tc_->frame_->size)
		tc_->frame_->size = tc_->frame_->tmpoff;
	return l;
}

// Given a variable v, return the corresponding loc.
Loc*
var2loc(Compile *cs, Var *v)
{
	int i;
	Loc *loc;
	Frame *f;

	if(v == nil)
		panic("var2loc");

	loc = cs->localloc(LocGlobal, v->type);
	if(v->f == globalframe){
		loc->mode = LocGlobal;
	}else if(v->escapes == 0){
		if(cs->fn_ == nil || v->f != cs->fn_->frame)
			panic("var2loc %p %s %T - bad frame - fn %p - frame %p %p", v, v->name, v->type, cs->fn_, v->f, cs->fn_ ? cs->fn_->frame : nil);
		loc->mode = LocFrame;
	}else{
		loc->mode = LocEscape;
		loc->depth = cs->tc_->frame_->depth - v->f->depth;
		f = cs->tc_->frame_;
		for(i=0; i<loc->depth; i++){
			if(f == nil || f == v->f)
				panic("var2loc - long depth");
			f = f->outer;
		}
		if(f != v->f)
			panic("var2loc - short depth");
	}
	loc->off = v->off;
	return loc;
}

// Given the location l, which must contain a pointer,
// return a location corresponding to the pointer itself.
Loc*
Compile::deref(Line line, Loc *l)
{
	Loc *l1, *tmp;

	if(l->type == nil || (l->type->op != TypePtr && l->type->op != TypeList))
		panic("deref %T", l->type);
	if(l->mode&LocIndirect){
		tmp = mktmp(l->type);
		emitmov(line, tmp, l);
		l = tmp;
	}
	l1 = localloc(l->mode|LocIndirect, l->type->op == TypePtr ? l->type->left : l->type);
	l1->off = l->off;
	l1->off1 = 0;
	l1->depth = l->depth;
	return l1;
}

// Save the current state of the temporary stack.
void
Compile::pushtmp()
{
	tmpstack_ = listinsert(&arena_, tmpstack_, tc_->frame_->tmpoff);
}

// Restore the temporary stack to what it was at the corresponding pushtmp.
// All slots created with mktmp since the last pushtmp are no longer valid.
void
Compile::poptmp()
{
	if(tmpstack_ == nil)
		panic("poptmp");
	tc_->frame_->tmpoff = tmpstack_->x;
	tmpstack_ = listpop(tmpstack_);
}

// Determine whether poptmp will invalidate loc.
int
popinvalidates(Compile *cs, Loc *loc)
{
	int o;

	if(cs->tmpstack_ == nil)
		panic("popinvalidates");
	if(loc == nil)
		return 0;
	if(cs->tc_->frame_ == globalframe){
		if(loc->mode != LocGlobal && loc->mode != LocGlobalIndirect)
			return 0;
	}else{
		if(loc->mode != LocFrame && loc->mode != LocFrameIndirect)
			return 0;
	}
	o = cs->tmpstack_->x;
	return loc->off >= o;
}

// Ensure the the codebuf b has room for n bytes.
void
bensure(Codebuf *b, int n)
{
	uchar *obp;
	int sz;

	if(b->p + n < b->ep)
		return;
	n += b->ep - b->bp;
	if(b->ep == b->bp)
		sz = 50;  // linkfn needs 44
	else
		sz = b->ep - b->bp;
	while(sz < n)
		sz += (sz>>1);
	obp = b->bp;
	b->bp = (uchar*)erealloc(b->bp, sz);
	b->ep = b->bp + sz;
	b->p += b->bp - obp;
}

// Ensure that the compile state cs's codebuf has room for n bytes.
void
Compile::ensure(int n)
{
	Codebuf *b = codebuf_;
	uchar *obp = b->bp;
	bensure(b, n);
	if(obp == b->bp)
		return;
	
	// If the code buf was reallocated, adjust all the pointers
	// stored in the labels we've accumulated.
	for(Label *l=tc_->labels_.next; l != (Label*)&tc_->labels_; l=l->next){
		if(l->val)
			l->val += b->bp - obp;
		for(List<int*> *ll=l->uses; ll; ll=ll->next)
			ll->x = (int*)((char*)ll->x + (b->bp - obp));
	}
}

// Now that we're at the end of the compilation,
// resolve references to labels.
void
patch(Compile *cs)
{
	for(Label *l=cs->tc_->labels_.next; l != (Label*)&cs->tc_->labels_; l=l->next){
		if(l->val == nil && l->uses == nil)
			error(ZL, "label with no def, no uses");	 // wow!
		else if(l->val == nil){
			if(l->used)
				error(l->used->line_, "label %s not defined [%p]", l->name, l);
		}
		else if(l->uses == nil && l->name){
			warn(ZL, "label %s not used", l->name);
		}
		else
			for(List<int*> *ll=l->uses; ll; ll=ll->next)
				*ll->x = l->val - (char*)(ll->x+1);
	}
}

// Allocate a code buffer.
Codebuf*
mkcodebuf(void)
{
	Codebuf *b;
	
	b = (Codebuf*)emalloc(sizeof(Codebuf));
	b->labels.next = b->labels.prev = (Label*)&b->labels;
	b->live = 1;
	bensure(b, 4);
	return b;
}

// Emit the pointer p to the code buffer.
void
Compile::emitp(void *p)
{
	ensure(sizeof(p));
	*codebuf_->vp++ = p;
}

// Emit the integer i to the code buffer.
void
Compile::emiti(int i)
{
	ensure(sizeof(i));
	*codebuf_->ip++ = i;
}

// Emit the floating-point number f to the code buffer.
void
Compile::emitf(double f)
{
	ensure(sizeof(f));
	*codebuf_->fp++ = f;
}

void
Compile::emitline(Line line)
{
	emiti((file2num(line.file)<<16) | line.line);
}

void
Compile::emitC(Line line, Code *code)
{
	emitline(line);
	emitp((void*)code);
}

// Emit an encoding of the location l to the code buffer.
void
Compile::emitloc(Loc *l)
{
	if(l == nil){
		emiti(~0);
		return;
	}
	if((l->mode&~LocIndirect) == LocEscape)
		emiti((l->depth<<3)|(l->off<<11)|l->mode);
	else
		emiti((l->off<<3)|l->mode);
	if(l->mode&LocIndirect)
		emiti(l->off1);
}

// Emit sequence of bytecode pieces. 
// Arguments encoded in function name:
//	C - code
//	P - pointer
//	L - loc
// XXX this is just begging to be autogenerated by Perl :)

void
Compile::emitCP(Line line, Code *code, void *p)
{
	emitC(line, code);
	emitp(p);
}

void
Compile::emitCL(Line line, Code *code, Loc *l1)
{
	emitC(line, code);
	emitloc(l1);
}

void
Compile::emitCLI(Line line, Code *code, Loc *l1, int i)
{
	emitC(line, code);
	emitloc(l1);
	emiti(i);
}

void
Compile::emitCLF(Line line, Code *code, Loc *l1, double f)
{
	emitC(line, code);
	emitloc(l1);
	emitf(f);
}

void
Compile::emitCLP(Line line, Code *code, Loc *l1, void *p)
{
	emitC(line, code);
	emitloc(l1);
	emitp(p);
}

void
Compile::emitCLL(Line line, Code *code, Loc *l1, Loc *l2)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
}

void
Compile::emitCLLP(Line line, Code *code, Loc *l1, Loc *l2, void *p)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
	emitp(p);
}

void
Compile::emitCLPL(Line line, Code *code, Loc *l1, void *p, Loc *l2)
{
	emitC(line, code);
	emitloc(l1);
	emitp(p);
	emitloc(l2);
}

void
Compile::emitCLLPP(Line line, Code *code, Loc *l1, Loc *l2, void *p1, void *p2)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
	emitp(p1);
	emitp(p2);
}

void
Compile::emitCPLL(Line line, Code *code, void *p, Loc *l1, Loc *l2)
{
	emitC(line, code);
	emitp(p);
	emitloc(l1);
	emitloc(l2);
}

void
Compile::emitCLLL(Line line, Code *code, Loc *l1, Loc *l2, Loc *l3)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
	emitloc(l3);
}

void
Compile::emitCPLLL(Line line, Code *code, void *p, Loc *l1, Loc *l2, Loc *l3)
{
	emitC(line, code);
	emitp(p);
	emitloc(l1);
	emitloc(l2);
	emitloc(l3);
}

void
Compile::emitCLLLL(Line line, Code *code, Loc *l1, Loc *l2, Loc *l3, Loc *l4)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
	emitloc(l3);
	emitloc(l4);
}

void
Compile::emitCLLLP(Line line, Code *code, Loc *l1, Loc *l2, Loc *l3, void *p)
{
	emitC(line, code);
	emitloc(l1);
	emitloc(l2);
	emitloc(l3);
	emitp(p);
}

void
Compile::emitCLPLL(Line line, Code *code, Loc *l1, void *p, Loc *l2, Loc *l3)
{
	emitC(line, code);
	emitloc(l1);
	emitp(p);
	emitloc(l2);
	emitloc(l3);
}

// Emit a jump (specified by code) to the code buffer.
// Dst is the target label for the jump.
// If code != xjmp, src is the value to use in a conditional jump.
// If code == xjmp, src is ignored.
void
Compile::emitjmp(Line line, Code *code, Loc *src, Label *dst)
{
	emitC(line, code);
	if(code != xjmp)
		emitloc(src);
	emiti(0);
	dst->uses = listappend(&arena_, dst->uses, codebuf_->ip-1);
}

// Emit a label to the code buffer.
// Doesn't actually emit any code; records the current code
// position to use when resolving references.
void
Compile::emitlabel(Label *l)
{
	if(l->val)
		panic("emitlabel");
	l->val = (char*)codebuf_->p;
}

// Emit code to load the pointer v into the given location.
void
Compile::emitloadptr(Line line, Loc *loc, void *v)
{
	if(loc)
		emitCLP(line, xloadp, loc, v);
}

// Emit code to load the intger i into the given location.
void
Compile::emitloadint(Line line, Loc *loc, int i)
{
	if(loc)
		emitCLI(line, xloadi, loc, i);
}

// Emit code to load the floating-point number f into the given location.
void
Compile::emitloadfloat(Line line, Loc *loc, double f)
{
	if(loc)
		emitCLF(line, xloadf, loc, f);
}

// Return a location offset from l by n bytes.
Loc*
Compile::offset(Loc *l, int n)
{
	Loc *tmp;
	
	tmp = mktmp(l->type);
	*tmp = *l;
	if(tmp->mode&LocIndirect)
		tmp->off1 += n;
	else
		tmp->off += n;
	return tmp;
}

// Emit code to copy the data at src into dst.
// src->type must give the type of the data being copied,
// to derive the size of the data as well as any special
// properties (like being an extensible or discriminated struct).
void
Compile::emitmov(Line line, Loc *dst, Loc *src)
{
	Type *t;
	int n;

	if(dst == nil)
		return;
	assert(dst->type);
	assert(src->type);
	t = src->type;
	// Discriminated and extensible structs must be copied specially.
	if(t->op == TypeStruct && (rootstruct(t)->discriminated || t->extensible)){
		assert(dst->type->op == TypeStruct);
		assert(rootstruct(dst->type) == rootstruct(t));
		emitCLL(line, xcopystruct, dst, src);
		emitp(dst->type);
		emitp(src->type);
	}
	// Other objects are copied with a simple memmove.
	else{
		n = sizeoftype(dst->type);
		assert(n <= sizeoftype(t));
		if(n == 4)
			emitC(line, xmov4);
		else{
			emitC(line, xmovn);
			emiti(n);
		}
		emitloc(dst);
		emitloc(src);
	}
}

// Emit code to write a pointer to fn's code into the destination dst.
void
Compile::emitloadfn(Line line, Loc *dst, Fn *fn)
{
	Loc *loc;

	if(fn->frame->escsize && fn->frame->outer != globalframe){
		// Function pointer with upward fnargs: create
		// new code snippet that loads the escape frame pointer
		// appropriately.
		loc = localloc(LocFrame, typeint()); // XXX typeptr()
		loc->off = FrameOffsetEP;
		emitCLLP(line, xlinkfn, dst, loc, fn);
	}else
		emitloadptr(line, dst, fn);
}

// Compile and emit code for a statement.
void
compilestmt(Compile *cs, Stmt *stmt)
{
	List<int> *starttmp = cs->tmpstack_;
	if(stmt == nil)
		return;

	while(stmt->expanded_)
		stmt = stmt->expanded_;
	stmt->compile(cs);
	assert(cs->tmpstack_ == starttmp);
}

void
StmtExpr::compile(Compile *cs)
{
	cs->pushtmp();
	compileexpr(cs, expr_);
	cs->poptmp();
}

void
StmtLoop::compile(Compile *cs)
{
	compileexpr(cs, pre_);
	if(!dowhile_ && cond_){
		cs->pushtmp();
		Loc *tmp = cs->evaltmp(cond_);
		cs->emitjmp(line_, xncjmp, tmp, lbreak_);
		cs->poptmp();
	}
	Label *l1 = mklabel(cs->tc_);
	cs->emitlabel(l1);
	compilestmt(cs, body_);
	cs->emitlabel(lcontinue_);
	compileexpr(cs, post_);
	if(cond_){
		cs->pushtmp();
		Loc *tmp = cs->evaltmp(cond_);
		cs->emitjmp(line_, xcjmp, tmp, l1);
		cs->poptmp();
	}else
		cs->emitjmp(line_, xjmp, nil, l1);
	cs->emitlabel(lbreak_);
}

void
StmtIf::compile(Compile *cs)
{
	cs->pushtmp();
	Loc *tmp = cs->evaltmp(cond_);
	Label *l1 = mklabel(cs->tc_);
	cs->emitjmp(cond_->line_, xncjmp, tmp, l1);
	cs->poptmp();
	compilestmt(cs, left_);
	if(right_){
		Label *l2 = mklabel(cs->tc_);
		cs->emitjmp(cond_->line_, xjmp, nil, l2);
		cs->emitlabel(l1);
		compilestmt(cs, right_);
		cs->emitlabel(l2);
	}else
		cs->emitlabel(l1);
}

void
StmtBreak::compile(Compile *cs)
{
	cs->emitjmp(line_, xjmp, nil, label_);
}

void
StmtContinue::compile(Compile *cs)
{
	cs->emitjmp(line_, xjmp, nil, label_);
}

void
StmtList::compile(Compile *cs)
{
	for(List<Stmt*> *l=stmts_; l; l=l->next)
		compilestmt(cs, l->x);
}

void
StmtPrint::compile(Compile *cs)
{
	cs->pushtmp();
	Loc *tmp = cs->evaltmp(expr_);
	cs->emitCP(line_, xprint, expr_->type_);
	cs->emitloc(tmp);
	cs->poptmp();
}

void
StmtReturn::compile(Compile *cs)
{
	if(expr_){
		Loc *loc = cs->localloc(LocFrame, expr_->type_);
		loc->mode = LocFrame|LocIndirect;
		loc->off = -sizeoftype(typeptr(nil));
		if(FLAGS_tailcalls && expr_->exprcall())
			cs->tailcall_ = 1;
		cs->evalinto(expr_, loc);
		if(FLAGS_tailcalls && expr_->exprcall())
			return;
	}
	cs->emitC(line_, xret);
}

void
StmtGoto::compile(Compile *cs)
{
	cs->emitjmp(line_, xjmp, nil, label_);
}

void
StmtLabel::compile(Compile *cs)
{
	cs->emitlabel(label_);
}

void
StmtFndef::compile(Compile *cs)
{
	compilefn(line_, fn_);
	cs->emitloadfn(line_, var2loc(cs, var_), fn_);
}

void
StmtDeclare::compile(Compile *cs)
{
	if(var_)
		cs->emitCLP(line_, xzero, var2loc(cs, var_), type_);
}

void
StmtDeclareAttribute::compile(Compile *cs)
{
	if(var_)
		cs->emitCLP(line_, xzero, var2loc(cs, var_), type_);
}

void
StmtDeclareType::compile(Compile *cs)
{
}

void
StmtStruct::compile(Compile *cs)
{
}

void
StmtCase::compile(Compile *cs)
{
	cs->pushtmp();
	{
		int canfail = 1, istmp;

		Loc *tmp = maybelevalexpr(cs, cond_, 0);
		if (tmp) {
			istmp = 0;
		} else {
			tmp = cs->evaltmp(cond_);
			istmp = 1;
		}

		// The label at the very end
		Label *l1 = mklabel(cs->tc_);
		for(List<Clause*> *l=clauses_; l; l=l->next){
			Clause *c = l->x;
			if(!canfail){
				// XXX warning
				error(c->pat->line, "unreachable clause");
			}
			// Pattern fail label
			Label *l2 = mklabel(cs->tc_);
			// Test the pattern
			canfail = 0;
			compilepattern(cs, c->pat, tmp, l2, &canfail, istmp);
			// Run the body
			compilestmt(cs, c->body);
			// Skip over the rest of the clauses
			if(l->next)
				cs->emitjmp(line_, xjmp, nil, l1);
			cs->emitlabel(l2);
		}
		// All done
		cs->emitlabel(l1);
	}
	cs->poptmp();
}

// Compile and emit code for an expression.
// The value is thrown away, so the expression should have side effects.
void
compileexpr(Compile *cs, Expr *expr)
{
	if(expr == nil)
		return;
	cs->evalexpr(expr, nil, nil);
}

// Compile a sequence of expressions, emitting their results in a
// packed tuple, starting at dst.
void
compiletuple(Compile *cs, const vector<Expr*>& exprs, Loc *dst)
{
	if(dst == nil){
		for(uint i=0; i<exprs.size(); i++)
			compileexpr(cs, exprs[i]);
	} else {
		Loc locx = *dst;
		Loc *loc = &locx;
		for(uint i=0; i<exprs.size(); i++){
			Expr *e = exprs[i];
			loc->type = e->type_;
			cs->evalinto(e, loc);
			loc = cs->offset(loc, sizeoftype(e->type_));
		}
	}
}

// Load effective address of expr into dst.
void
Compile::leaexpr(Expr *expr, Loc *dst)
{
	Loc *loc;

	loc = levalexpr(this, expr, 1);
	emitCLL(expr->line_, xlea, dst, loc);
}

// Convert boolean src into dst.
Loc*
cvtbool(Compile *cs, Line line, Loc *src)
{
	Loc *dst;
	Code *cvt;

	if(src->type == nil)
		panic("cvtbool2");

	dst = cs->mktmp(typebool());
	switch(src->type->op){
	case TypeList:
	case TypePtr:
	case TypeInt:
	case TypeBool:
		cvt = xmov4;
		break;

	case TypeFloat:
		cvt = xnzf;
		break;

	case TypeString:
		cvt = xnzs;
		break;

	default:
		cvt = nil;
		error(ZL, "cannot convert %T for jump", src->type);
		break;
	}
	cs->emitCLL(line, cvt, dst, src);
	return dst;
}

// Evaluate expr into loc. 
// (Result must go into loc & loc can be scribbled on during eval.)
Loc*
Compile::evalinto(Expr *expr, Loc *loc)
{
	Loc *tmp;

	tmp = evalexpr(expr, loc, loc);
	if(tmp != loc)
		emitmov(expr->line_, loc, tmp);
	return loc;
}

// Evaluate expr, possibly into a new temporary location.
// (Can go away when caller does poptmp!)
Loc*
Compile::evaltmp(Expr *expr)
{
	Loc *tmp;
	
	if(expr == nil)
		return nil;
	tmp = mktmp(expr->type_);
	return evalexpr(expr, tmp, tmp);
}

// Construct integer literal.  Goes away at next poptmp.
Loc*
Compile::evalint(Line line, int n)
{
	Loc *tmp;
	
	tmp = mktmp(typeint());
	emitloadint(line, tmp, n);
	return tmp;
}

// Construct floating literal.  Goes away at next poptmp.
Loc*
Compile::evalfloat(Line line, double f)
{
	Loc *tmp;
	
	tmp = mktmp(typefloat());
	emitloadfloat(line, tmp, f);
	return tmp;
}

// Construct list literal, saving head in head.
Loc*
evalmklist(Compile *cs, Line line, Type *t, const vector<Expr*>& exprs, Loc *head)
{
	Loc *tailp, *cons, *loc;

	if(head == nil){
		for(uint i=0; i<exprs.size(); i++)
			compileexpr(cs, exprs[i]);
		return nil;
	}
	
	cs->emitloadptr(line, head, nil);

	if(exprs.size() == 0)
		return head;

	// tailp = &head;
	tailp = cs->mktmp(typeptr(t));
	cs->emitCLL(line, xlea, tailp, head);
	cons = cs->mktmp(t);
	for(uint i=0; i<exprs.size(); i++){
		// cons = new ListNode;
		cs->emitCLP(line, xnew, cons, t);
		
		// cons->hd = l->x;
		loc = cs->offset(cs->deref(line, cons), offsetof(ListNode, v));
		loc->type = t->left;
		cs->evalinto(exprs[i], loc);
		
		// *tailp = cons;
		cs->emitmov(line, cs->deref(line, tailp), cons);
		
		// tailp = &cons->tl (= cons)
		cs->emitCLL(line, xmov4, tailp, cons);
	}

	return head;
}

// Construct literal array, saving pointer in array.
Loc*
evalmkarray(Compile *cs, Line line, Type *t, const vector<Expr*>& exprs, Loc *array)
{
	int n, step;
	Loc *elem;

	n = exprs.size();
	if(array == nil){
		for(int i=0; i<n; i++)
			compileexpr(cs, exprs[i]);
		return nil;
	}
	
	// Allocate array of size n.
	cs->pushtmp();
	cs->emitCLLP(line, xnewa, array, cs->evalint(line, n), t->left);
	cs->poptmp();

	if(n == 0)
		return array;
	
	// elem = &array[0];
	elem = cs->mktmp(typeptr(t->left));
	cs->pushtmp();
	cs->emitCLLL(line, xleaa, elem, array, cs->evalint(line, 0));
	cs->poptmp();
	
	// Populate array.
	step = sizeoftype(t->left);
	for(int i=0; i<n; i++)
		cs->evalinto(exprs[i], cs->offset(cs->deref(line, elem), i*step));

	return array;
}

// Construct literal map, saving pointer in map.
Loc*
evalmkmap(Compile *cs, Line line, Type *t, const vector<ExprMap*>& exprs, Loc *map)
{
	int n = exprs.size();
	if(map == nil){
		for(int i=0; i<n; i++)
			compileexpr(cs, exprs[i]);
		return nil;
	}

	// Allocate map.
	cs->emitCLP(line, xnew, map, t);

	// Loop over constructors, adding to map.
	for(int i=0; i<n; i++){
		ExprMap *e = exprs[i];
		cs->pushtmp();
		// valp = &m[key];
		Loc *valp = cs->mktmp(typeptr(t->right));
		cs->emitCLLL(line, xleam, valp, map, cs->evaltmp(e->left_));
		// *valp = value
		cs->evalinto(e->right_, cs->deref(line, valp));
		cs->poptmp();
	}
	
	return map;
}

// Construct cons cell, saving pointer in cons.
Loc*
evalmkcons(Compile *cs, Line line, Type *t, Expr *left, Expr *right, Loc *cons)
{
	Loc *loc;

	if(cons == nil){
		compileexpr(cs, left);
		compileexpr(cs, right);
		return nil;
	}

	cs->emitCLP(line, xnew, cons, t);
	// cons->hd = left
	loc = cs->offset(cs->deref(line, cons), offsetof(ListNode, v));
	loc->type = left->type_;
	cs->evalinto(left, loc);

	// cons->tl = right
	loc = cs->offset(cs->deref(line, cons), 0);
	loc->type = right->type_;
	cs->evalinto(right, loc);

	return cons;
}

// Convert e into type t, saving in dst.
Loc*
evalconvert(Compile *cs, Line line, Type *t, Expr *e, Loc *dst)
{
	Code *cvt;
	Loc *tmp;

	if(dst == nil){
		compileexpr(cs, e);
		return nil;
	}

	if(e->type_ == t || e->type_ == nil){
	NoConversion:
		cs->evalinto(e, dst);
		return dst;
	}
	
	cvt = nil;
	switch(t->op){
	case TypeBool:
		switch(e->type_->op){
		case TypeInt:
		case TypePtr:
		case TypeList:
		case TypeFn:
		case TypeMap:
		case TypeForall:
			goto NoConversion;
			break;
		case TypeFloat:
			cvt = xnzf;
			break;
		case TypeString:
			cvt = xnzs;
			break;
		default:;
		}
		break;
	
	case TypeInt:
		switch(e->type_->op){
		case TypeBool:
			goto NoConversion;
			break;
		case TypeFloat:
			cvt = xftoi;
			break;
		case TypeString:
			cvt = xatoi;
			break;
		default:;
		}
		break;
	
	case TypeFloat:
		switch(e->type_->op){
		case TypeBool:
			cvt = xbtof;
			break;
		case TypeInt:
			cvt = xitof;
			break;
		case TypeString:
			cvt = xatof;
			break;
		default:;
		}
		break;
	
	case TypeString:
		switch(e->type_->op){
		case TypeBool:
			cvt = xbtoa;
			break;
		case TypeInt:
			cvt = xitoa;
			break;
		case TypeFloat:
			cvt = xftoa;
			break;
		default:
			if(isptrast(e->type_))
				cvt = xasttoa;
			break;
		}
		break;
	
	case TypeList:
		if(isptrast(e->type_))
			cvt = xasttolist;
		else if(e->type_->op == TypeArray)
			cvt = xarraytolist;
		break;

	case TypeArray:
		if(e->type_ == typeint()){
			// Must be able to compute the size of this array
			assert(t->left);
			// New array of size e
			cs->pushtmp();
			cs->emitCLLP(line, xnewa, dst, cs->evaltmp(e), t->left);
			cs->poptmp();
			return dst;
		}
		else if(e->type_->op == TypeList)
			cvt = xlisttoarray;
		break;
	
	case TypeTuple:
		// MapIter --> Tuple
		// XXX need more checking here?
		cvt = xmapitertuple;
		break;
	
	case TypeMapIter:
		// Map --> MapIter
		// XXX need more checking here?
		cvt = xnewmapiter;
		break;
	
	default:
		if(isptrast(t)){
			// This is an option() conversion for ASTs
			cvt = xasttooption;
		}
		break;
	}
	
	// Emit the conversion
	if(cvt){
		cs->pushtmp();
		tmp = cs->evaltmp(e);
		if(cvt == xlisttoarray || cvt == xarraytolist)
			cs->emitCPLL(line, cvt, dst->type->left, dst, tmp);
		else
			cs->emitCLL(line, cvt, dst, tmp);
		cs->poptmp();
	}else
		error(e->line_, "cannot convert %T to %T", e->type_, t);
	return dst;
}

// Evaluate expr, returning a pointer to the location storing expr.
// Can use the location dstatomic to store expr, except that it's
// contents might be needed to compute expr, so don't smash it
// too early.
// Can use the location dsttmp to store expr or to store temporary
// contents -- it is empty and can be scribbled on at will.
//
// Dstatomic is the eventual place where the expr will be stored,
// so if it doesn't require any extra instructions, try to store it there.
// If dstatomic == nil, then the value is not needed -- just evaluate
// for side-effects.
Loc*
Compile::evalexpr(Expr *expr, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out;

	if(expr == nil){
		error(ZL, "missing expr");
		return nil;
	}

	while(expr->expanded_)
		expr = expr->expanded_;

	if(expr->type_ == nil)
		return nil;

	pushtmp();
	if(dstatomic && dsttmp == nil)
		dsttmp = mktmp(dstatomic->type);

	out = expr->compile(this, dstatomic, dsttmp);

	if(dstatomic){
		if(out == nil)
			panic("evalexpr %E - no out", expr);
		if(popinvalidates(this, out)){
			emitmov(expr->line_, dstatomic, out);
			out = dstatomic;
		}
	}
	poptmp();
	return out;
}

Loc*
Expr::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	panic("expr::compile - %#E", this);
}

Loc*
ExprCode::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(assign_){
		Loc *src = levalexpr(cs, left_, 1);
		Loc *tmp = cs->evaltmp(right_);
		cs->emitCLLL(line_, code_, src, src, tmp);
		return src;
	}
	if(dstatomic == nil){
		compileexpr(cs, left_);
		if(right_)
			compileexpr(cs, right_);
		return nil;
	}
	Loc *tmp = cs->evaltmp(left_);
	if(right_){
		Loc *tmp1 = cs->evaltmp(right_);
		cs->emitCLLL(line_, code_, dstatomic, tmp, tmp1);
	}else
		cs->emitCLL(line_, code_, dstatomic, tmp);
	return dstatomic;
}

Loc*
ExprAdd::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(type_->op == TypeList){
		Loc *tmp = cs->evaltmp(left_);
		Loc *tmp1 = cs->evaltmp(right_);
		if(dstatomic == nil)
			return nil;
		cs->emitCPLLL(line_, xaddl, type_->left, dstatomic, tmp, tmp1);
		return dstatomic;
	}
	return ExprCode::compile(cs, dstatomic, dsttmp);
}

Loc*
ExprAddEq::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(type_->op == TypeList){
		Loc *src = levalexpr(cs, left_, 1);
		Loc *tmp = cs->evaltmp(right_);
		cs->emitCPLLL(line_, xaddl, type_->left, src, src, tmp);
		return src;
	}
	return ExprCode::compile(cs, dstatomic, dsttmp);
}

Loc*
ExprIncDec::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *src = levalexpr(cs, left_, 1);
	if(!post_ && inc_ == 1 && left_->type_->op == TypeMapIter){
		cs->emitCLL(line_, xmapiternext, dstatomic, src);
		return dstatomic;
	}
	if(post_)
		cs->emitmov(line_, dsttmp, src);
	if(left_->type_ == typeint())
		cs->emitCLLL(line_, xaddi, src, src, cs->evalint(line_, inc_));
	else
		cs->emitCLLL(line_, xaddf, src, src, cs->evalfloat(line_, inc_));
	if(!post_){
		cs->emitmov(line_, dstatomic, src);
		return dstatomic;
	}
	return dsttmp;
}

Loc*
compileeq(Compile *cs, Line line, Expr *left, Expr *right, Loc *dstatomic, Loc *dsttmp, int neg)
{
	if(dstatomic == nil){
		compileexpr(cs, left);
		compileexpr(cs, right);
		return nil;
	}
	Loc *tmp = cs->evaltmp(left);
	Loc *tmp1 = cs->evaltmp(right);
	Code *code;
	if(left->type_->op == TypeFloat)
		code = neg ? xnef : xeqf;
	else
		code = neg ? xnei : xeqi;
	cs->emitCLLL(line, code, dstatomic, tmp, tmp1);
	return dstatomic;
}

Loc*
ExprEqEq::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return compileeq(cs, line_, left_, right_, dstatomic, dsttmp, 0);
}

Loc*
ExprNotEq::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return compileeq(cs, line_, left_, right_, dstatomic, dsttmp, 1);
}

Loc*
ExprDot::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return levalexpr(cs, this, 0);
}

Loc*
ExprName::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return levalexpr(cs, this, 0);
}

Loc*
ExprStar::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return levalexpr(cs, this, 0);
}

Loc*
ExprArray::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return levalexpr(cs, this, 0);
}

Loc*
ExprMkList::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	evalmklist(cs, line_, type_, exprs_, dsttmp);
	return dsttmp;
}

Loc*
ExprMkArray::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	evalmkarray(cs, line_, type_, exprs_, dsttmp);
	return dsttmp;
}

Loc*
ExprMkMap::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	evalmkmap(cs, line_, type_, exprs_, dsttmp);
	return dsttmp;
}

Loc*
ExprListCons::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	evalmkcons(cs, line_, type_, left_, right_, dsttmp);
	return dsttmp;
}

Loc*
ExprAndAnd::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out;

	Label *l1 = mklabel(cs->tc_);
	if(dstatomic == nil){
		out = nil;
		// Only care about side-effects.
		cs->pushtmp();
		cs->emitjmp(line_, xncjmp, cvtbool(cs, line_, cs->evaltmp(left_)), l1);
		cs->poptmp();
		compileexpr(cs, right_);
		cs->emitlabel(l1);
	}else{
		// Evaluate left into temporary.
		cs->evalinto(left_, out = dsttmp);
	
		// if(left isn't good) evaluate right into temporary.
		cs->emitjmp(line_, xncjmp, cvtbool(cs, line_, dsttmp), l1);
		cs->evalinto(right_, dsttmp);
		cs->emitlabel(l1);
	}
	return out;
}
Loc*
ExprOrOr::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	Label *l1 = mklabel(cs->tc_);
	if(dstatomic == nil){
		// Only care about side-effects.
		cs->pushtmp();
		cs->emitjmp(line_, xcjmp, cvtbool(cs, line_, cs->evaltmp(left_)), l1);
		cs->poptmp();
		compileexpr(cs, right_);
		cs->emitlabel(l1);
	}else{
		// Evaluate left into temporary.
		cs->evalinto(left_, out = dsttmp);
	
		// if(left isn't good) evaluate right into temporary.
		cs->emitjmp(line_, xcjmp, cvtbool(cs, line_, dsttmp), l1);
		cs->evalinto(right_, dsttmp);
		cs->emitlabel(l1);
	}
	return out;
}


static Loc*
compileassign(Compile *cs, Line line, Expr *left, Expr *right, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out;

	if((out = simpleassign(cs, left)) != nil){
		Loc *tmp = cs->evalexpr(right, out, dsttmp);
		if(tmp != out)
			cs->emitmov(line, out, tmp);
		if(tmp == dstatomic)
			out = tmp;
	}
	else if(dstatomic == nil)
		assign(cs, line, left, cs->evaltmp(right));
	else
		assign(cs, line, left, cs->evalinto(right, out = dsttmp));
	return out;
}

Loc*
ExprEq::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return compileassign(cs, line_, left_, right_, dstatomic, dsttmp);
}

Loc*
ExprDeclEq::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	return compileassign(cs, line_, left_, right_, dstatomic, dsttmp);
}

Loc*
ExprTuple::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	compiletuple(cs, exprs_, dsttmp);
	return dsttmp;
}

Loc*
ExprCall::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	// Special case (sorry)
	ExprName *en = left_->exprname();
	if(en && strcmp(en->name_, "listmap") == 0){
		assert(xtype_);
		assert(xtype_->op == TypeTuple);
		Loc *tmp1 = cs->evaltmp(exprs_[0]);
		Loc *tmp2 = cs->evaltmp(exprs_[1]);
		cs->emitCPLLL(line_, xlistmap, xtype_, dstatomic, tmp1, tmp2);
		return dstatomic;
	}

	// Normal call
	int tailcall = cs->tailcall_;
	cs->tailcall_ = 0;
	Loc *out = nil;
	Loc *loc, *tmp, *tmp1;
	Type *t;
	int n;
	switch(callop_){
	case CallRegular:
		if(left_ && left_->exprname())
			tmp = levalexpr(cs, left_, 0);
		else
			tmp = cs->evaltmp(left_);
		tmp1 = cs->mktmp(typeint());  // XXX typeptr()
		cs->emitCLL(line_, xmkframe, tmp1, tmp);
		loc = cs->localloc(tmp1->mode|LocIndirect, nil);
		loc->off = tmp1->off;
		loc->off1 = 0;
		loc->off1 -= sizeoftype(typeptr(nil));	/* return pointer */
		out = dsttmp;
		if(dsttmp == nil)
			dsttmp = cs->mktmp(type_);
		cs->emitCLL(line_, xlea, loc, dsttmp);
		for(uint j=0; j<exprs_.size(); j++){
			Expr *e = exprs_[j];
			loc->off1 -= sizeoftype(e->type_);
			loc->type = e->type_;
			cs->evalinto(e, loc);
		}
		cs->emitCLL(line_, tailcall ? xretcall : xcall, tmp, tmp1);
		
		/*
		 * this no-op is here so that the line of the instruction
		 * after the call is line_, so that stack traces look good.
		 */
		cs->emitC(line_, xnop);

		break;
	case CallCtor:
		t = type_;
		if(t->op == TypeSubst)
			t = t->left;
		if(t->op == TypeForall)
			t = t->left;
		assert(t->op == TypeStruct);
		// Construct the fields if the ctor has arguments
		if(exprs_.size() > 0){
			n = structdataoffset(t);
			compiletuple(cs, exprs_, cs->offset(out = dsttmp, n));
			// Zero the reserved and bookkeeping data blocks
			cs->emitCLI(line_, xzeron, out, n);
		}
		// With no arguments, just zero the struct
		else{
			cs->emitCLP(line_, xzero, out = dstatomic, t);
		}
		// Set the tag type ptr
		if(t->super)
			cs->emitloadptr(line_, cs->offset(out, tagtypeoffset(t)), t);
		break;
	}
	return out;
}

Loc*
ExprNil::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	if(dstatomic)
		cs->emitloadptr(line_, out = dstatomic, nil);
	return out;
}

Loc*
ExprBool::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	if(dstatomic)
		cs->emitloadint(line_, out = dstatomic, b_);
	return out;
}

Loc*
ExprInteger::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	if(dstatomic)
		cs->emitloadint(line_, out = dstatomic, i_);
	return out;
}

Loc*
ExprFloat::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	if(dstatomic)
		cs->emitloadfloat(line_, out = dstatomic, f_);
	return out;
}

Loc*
ExprString::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	if(dstatomic)
		cs->emitloadptr(line_, out = dstatomic, nameof(cunescapestring(str_, len_)));
	return out;
}

Loc*
ExprFn::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *out = nil;
	compilefn(line_, fn_);
	if(dstatomic)
		cs->emitloadfn(line_, out = dstatomic, fn_);
	return out;
}

Loc*
ExprNew::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil)
		return nil;
	Type *t = type_;
	if(left_->exprstar() && t->op == TypePtr && t->left->op == TypeStruct){
		// Special case: discriminated struct copying - new *(ptr struct)
		Loc *tmp = cs->evaltmp(left_->exprstar()->left_);
		cs->emitCLLP(line_, xnewcopy, dstatomic, tmp, t->left);
		return dstatomic;
	}
	cs->emitCLP(line_, xnew, dsttmp, left_->type_);
	cs->evalinto(left_, cs->deref(line_, dsttmp));
	return dsttmp;
}

Loc*
ExprInit::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	if(dstatomic == nil)
		return nil;
	Type *t = type_;
	// Zero the struct and set the tag type ptr
	cs->emitCLP(line_, xzero, dsttmp, t);
	if(t->super)
		cs->emitloadptr(line_, cs->offset(dsttmp, tagtypeoffset(t)), t);
	// Initialize the fields
	for(int j=0; j<exprs_.size(); j++){
		ExprInitField *e = exprs_[j];
		if(!e->type_)
			continue;
		Loc *loc = derefmember(cs, line_, dsttmp, t, e, 0);
		cs->evalinto(e->left_, loc);
	}
	return dsttmp;
}

Loc*
ExprConvert::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	evalconvert(cs, line_, type_, left_, dsttmp);
	return dsttmp;
}

Loc*
ExprAddr::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	cs->leaexpr(left_, dstatomic);
	return dstatomic;
}

Loc*
ExprLength::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *tmp = cs->evaltmp(left_);
	Type *t = left_->type_;
	if(t->op == TypeArray)
		cs->emitCLL(line_, xlengtha, dstatomic, tmp);
	else if(t->op == TypeList)
		cs->emitCLL(line_, xlengthl, dstatomic, tmp);
	else if(t->op == TypeString)
		cs->emitCLL(line_, xlengths, dstatomic, tmp);
	else
		panic("bad type for length()");
	return dstatomic;
}

Loc*
ExprReverse::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *tmp = cs->evaltmp(left_);
	if(dstatomic == nil)
		return nil;
	cs->emitCLLP(line_, xreversel, dstatomic, tmp, left_->type_);
	return dstatomic;
}

Loc*
ExprTag::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *tmp = cs->evaltmp(left_);
	if(dstatomic == nil)
		return nil;
	cs->emitCLLP(line_, xtagstring, dstatomic, tmp, rootstruct(left_->type_->left));
	return dstatomic;
}

Loc*
ExprSlice::compile(Compile *cs, Loc *dstatomic, Loc *dsttmp)
{
	Loc *tmp = cs->evaltmp(left_);
	Loc *tmp1 = cs->evaltmp(middle_);
	Loc *tmp2 = cs->evaltmp(right_);
	if(dstatomic == nil)
		return nil;
	if(type_->op == TypeString)
		cs->emitCLLLL(line_, xslices, dstatomic, tmp, tmp1, tmp2);
	else if(type_->op == TypeArray)
		cs->emitCLLLL(line_, xslicea, dstatomic, tmp, tmp1, tmp2);
	else
		panic("compile slice");
	return dstatomic;
}

// If e is a simple expression, return the location
// where an assign would write.
Loc*
simpleassign(Compile *cs, Expr *e)
{
	if(e == nil || e->type_ == nil || e->type_->op == TypeDontCare || e->exprtuple())
		return nil;
	return levalexpr(cs, e, 1);
}


// Assign the already-computed value stored at right
// to the (possibly-tupled) variables given by left. 
void
assign(Compile *cs, Line line, Expr *left, Loc *right)
{
	Loc loc;
	Expr *e;
	Type *t;
	int i;

	if(left == nil || right == nil)
		return;
	if(left->type_->op == TypeDontCare)
		return;
	cs->pushtmp();
	ExprTuple *et;
	if((et = left->exprtuple()) != nil){
		loc = *right;
		t = loc.type;
		assert(t->op == TypeTuple);
		List<Type*> *lt;
		for(i=0, lt=t->types; i<et->exprs_.size() && lt; i++, lt=lt->next){
			e = et->exprs_[i];
			loc.type = e->type_;
			assign(cs, line, e, &loc);
			if(loc.mode&LocIndirect)
				loc.off1 += sizeoftype(lt->x);
			else
				loc.off += sizeoftype(lt->x);
		}
	}else
		cs->emitmov(line, levalexpr(cs, left, 1), right);
	cs->poptmp();
	// TO DO: assignment of tuple to struct, vice versa.	
}

// Evaluate the address of an expression,
// returning a corresponding location.
Loc*
maybelevalexpr(Compile *cs, Expr *expr, int write)
{
	if(expr == nil)
		return nil;
	while(expr->expanded_)
		expr = expr->expanded_;
	return expr->lea(cs, write);
}

Loc*
levalexpr(Compile *cs, Expr *expr, int write)
{
	Loc *res;

	res = maybelevalexpr(cs, expr, write);
	if(!res)
		panic("levalexpr %E - %#E", expr, expr);
	return res;
}


Loc*
ExprStar::lea(Compile *cs, int write)
{
	Loc *tmp = cs->mktmp(left_->type_);
	cs->evalinto(left_, tmp);
	return cs->deref(line_, tmp);
}

Loc*
ExprName::lea(Compile *cs, int write)
{
	if(var_ == nil)
		return nil;
	return var2loc(cs, var_);
}

Loc*
ExprDot::lea(Compile *cs, int write)
{
	if(!left_->type_)
		return nil;
	int isptr = (left_->type_->op == TypePtr || left_->type_->op == TypeList);
	Type *tl;
	if(isptr)
		tl = left_->type_->left;
	else
		tl = left_->type_;
	Loc *lhs;
	if(isptr){
		lhs = cs->evaltmp(left_);
	}else{
		lhs = levalexpr(cs, left_, write);
		lhs->type = type_;
	}
	Loc *loc = derefmember(cs, line_, lhs, tl, this, isptr);
	return loc;
}

Loc*
ExprArray::lea(Compile *cs, int write)
{
	Type *tl = left_->type_;
	Type *tr = right_->type_;
	if(tl == nil || tr == nil)
		return nil;
	Loc *tmp = cs->evaltmp(left_);
	Loc *tmp1 = cs->evaltmp(right_);
	Loc *loc;
	if(tl->op == TypeString){
		if(write)
			error(line_, "cannot write to strings");
		loc = cs->mktmp(typestring());
		cs->emitCLL(line_, xstringindex, loc, tmp);
		cs->emitloc(tmp1);
	}else if(tl->op == TypeArray){
		loc = cs->mktmp(typeptr(type_));
		cs->emitCLL(line_, xleaa, loc, tmp);
		cs->emitloc(tmp1);
		loc = cs->deref(line_, loc);
	}else if(tl->op == TypeMap){
		loc = cs->mktmp(typeptr(type_));
		cs->emitCLL(line_, xleam, loc, tmp);
		cs->emitloc(tmp1);
		loc = cs->deref(line_, loc);
	}else if(tl->op == TypeList){
		if(isassoclist(tl, nil, nil)){
			loc = cs->mktmp(typeptr(type_));
			cs->emitCLL(line_, xleaassoc, loc, tmp);
			cs->emitloc(tmp1);
			loc = cs->deref(line_, loc);
		}else{
			loc = cs->mktmp(typeptr(type_));
			cs->emitCLL(line_, xlealistn, loc, tmp);
			cs->emitloc(tmp1);
			loc = cs->deref(line_, loc);
		}
	}else{
		error(line_, "bad array ref %E (%T) [%p %p]", this, tl, this, expanded_);
		loc = nil;
	}
	return loc;
}

// Return the location of a struct member.
Loc*
derefmember(Compile *cs, Line line, Loc *structloc, Type *structtype, Expr *expr, int isptr)
{
	Loc *loc, *tmp;
	int off;
	Var *var;
	Type *type;
	Type *xtype;
	
	ExprDot *ed;
	ExprInitField *eif;
	if((ed = expr->exprdot()) != nil){
		var = ed->var_;
		type = ed->type_;
		xtype = ed->xtype_;
	}else if((eif = expr->exprinitfield()) != nil){
		var = eif->var_;
		type = eif->type_;
		xtype = eif->xtype_;
	}else{
		assert(0);
		return nil;
	}

	// Is the field an attribute?
	if(var->attrfn || var->isbuiltinattr) {
		loc = cs->mktmp(type);
		cs->emitCLLP(line, xattr, loc, structloc, var);
		return loc;
	}

	if(var->host == typeast()->structrep && strcmp(var->name, "parent") == 0) {
		loc = cs->mktmp(type);
		cs->emitCLLP(line, xastparent, loc, structloc, nil);
		return loc;
	}

	if(xtype)
		// Construct location of extension pointer.
		off = extptroffset(structtype);
	else
		// Construct location of field.
		off = var->off;
		
	if(isptr)
		loc = cs->offset(cs->deref(line, structloc), off);
	else
		loc = cs->offset(structloc, off);
	loc->type = type;

	// Redirect through the extension pointer.
	if(xtype){
		tmp = cs->mktmp(typeptr(xtype));
		cs->emitCLL(line, xleax, tmp, loc);
		cs->emitp(xtype);
		loc = tmp;
		loc->type = type;
		loc->mode |= LocIndirect;
		loc->off1 = var->off;
	}
	return loc;
}

// Compile the code for a function.
void
compilefn(Line line, Fn *fn)
{
	Compile *cs;
	Stmt *s;
	Loc *loc, *loc1;
	Var *v;

	cs = new Compile(fn->tc);
	cs->fn_ = fn;
	cs->codebuf_ = mkcodebuf();
	fn->code = cs->codebuf_;
	s = nil;
	if(fn->frame->escsize){
		// Assume trampoline has initialized EP = outer EP
		// if there is an outer EP.  See emitloadfn above.

		// Allocate escape frame.
		cs->pushtmp();
		loc = cs->mktmp(typeint()); // XXX typeptr()
		cs->emitCLI(line, xalloc, loc, fn->frame->escsize);

		// local EP pointer
		loc1 = cs->localloc(LocFrame, typeint());  // XXX typeptr()
		loc1->off = FrameOffsetEP;

		// Copy outer EP into escape frame.
		if(fn->frame->outer && fn->frame->outer->escsize){
			loc->mode |= LocIndirect;
			cs->emitmov(line, loc, loc1);
			loc->mode &= ~LocIndirect;
		}

		// Copy escape frame into EP.
		cs->emitmov(line, loc1, loc);
		cs->poptmp();
		
		// Copy arguments into EP as needed.
		loc->mode = LocFrame;
		for(List<Var*> *l=fn->frame->vars; l; l=l->next){
			v = l->x;
			if(!v->escapes)
				continue;
			if(v->argoff <= 0){
				loc->off = v->argoff;
				loc->type = v->type;
				cs->emitmov(line, var2loc(cs, v), loc);
			}
		}	
	}

	for(List<Stmt*> *l=fn->stmts; l; l=l->next)
		compilestmt(cs, s=l->x);
	if(cs->tc_->reachable_){
		Line line1;
		line1 = fn->line;
		line1.line = line1.line1;
		cs->emitC(line1, xret);
		if(fn->rettype != typevoid())
			error(s ? s->line_ : line, "function missing return, last stmt: %#$", s);
	}
	patch(cs);
	aclear(&cs->arena_);
	if(FLAGS_disassemble){
		fprint(2, "--fn %p argsize=%d size=%d tmpoff=%d esc=%d %s\n", fn,
			fn->frame->argsize, fn->frame->size,
			fn->frame->tmpoff, fn->frame->escsize, fn->name);
		disasm(cs->codebuf_->bp, cs->codebuf_->p);
		fprint(2, "--\n");
	}
	
	delete cs;
	fn->cs = nil;
}

extern void emitpatternast(Compile*, ExprParse*);
extern void emitissubtype(Compile*, Line, Loc*, Loc*, Type*, int);

// Compile a case pattern.  canfail is set to 1 if the pattern can
// cause a jump to the failure label.  If istmp is 1, errors are
// generated for pointer literals.
void
compilepattern(Compile *cs, Pattern *pat, Loc *val, Label *fail,
	       int *canfail, int istmp)
{
	Loc *cmp, *tmp, *loc, loc1;	/* Temp vars */
	Loc fieldloc;
	Type *t, *tl, *tr;
	Expr *e;
	Label *subfail, *subsucceed;
	int subcanfail, subcanfail2;
	int ptrsz;

	// This happens when the pattern fails to typecheck.
	// It probably would be better to do this up at the clause level.
	if(pat->type == typevoid()){
		cs->emitjmp(pat->line, xjmp, nil, fail);
		*canfail = 1;
		return;
	}

	switch(pat->op){
	case PatternTag:
		assert(val->type->op == TypePtr);
		assert(val->type->left->op == TypeStruct
		    || isptrast(val->type));
		// Test for nil
		// TODO: check pat->tagtype
		cs->emitjmp(pat->line, xncjmp, val, fail);
		*canfail = 1;
		// Perform the subtype test, and jump if it failed.
		cs->pushtmp();
		cmp = cs->mktmp(typeint());
		emitissubtype(cs, pat->line, cmp, val, pat->tagtype, 0);
		cs->emitjmp(pat->line, xncjmp, cmp, fail);
		cs->poptmp();
		// "Typecast" the value location, now that we know this is safe
		loc1 = *val;
		loc1.type = typeptr(pat->tagtype);
		// Process the subpattern
		compilepattern(cs, pat->subpat, &loc1, fail, canfail, istmp);
		break;
	case PatternDtor:
		t = val->type;
		// Compute the location of the first field
		fieldloc = *val;
		if(t->op == TypePtr){
			t = t->left;
			fieldloc = *cs->deref(pat->line, &fieldloc);
		}
		if(t->op == TypeStruct)
			fieldloc = *cs->offset(&fieldloc, structdataoffset(t));
		// Match base sub-patterns
		List<Type*> *l;
		List<Pattern*> *lp;
		for(l=t->ctorargs, lp=pat->subpats; l && lp; l=l->next, lp=lp->next){
			Type *t = l->x;
			Pattern *subpat = lp->x;
			fieldloc.type = t;
			compilepattern(cs, subpat, &fieldloc, fail, canfail, istmp);
			fieldloc = *cs->offset(&fieldloc, sizeoftype(t));
		}
		break;
	case PatternEmptyList:
		assert(val->type->op == TypeList);
		// Check for an empty list
		cs->pushtmp();
		tmp = cs->mktmp(typeint());
		cmp = cs->mktmp(typeint());
		cs->emitloadptr(pat->line, tmp, nil);
		cs->emitCLLL(pat->line, xeqi, cmp, val, tmp);
		cs->emitjmp(pat->line, xncjmp, cmp, fail);
		*canfail = 1;
		cs->poptmp();
		break;
	case PatternList:
		assert(val->type->op == TypeList);
		ptrsz = sizeoftype(val->type);
		// Check for an empty list
		cs->pushtmp();
		tmp = cs->mktmp(typeint());
		cmp = cs->mktmp(typeint());
		cs->emitloadptr(pat->line, tmp, nil);
		cs->emitCLLL(pat->line, xeqi, cmp, val, tmp);
		cs->emitjmp(pat->line, xcjmp, cmp, fail);
		*canfail = 1;
		cs->poptmp();
		// Match the list head
		loc1 = *cs->offset(cs->deref(pat->line, val), ptrsz);
		loc1.type = pat->type->left;
		compilepattern(cs, pat->subpat, &loc1, fail, canfail, istmp);
		// Match the list tail
		loc1 = *cs->offset(cs->deref(pat->line, val), 0);
		loc1.type = pat->type;
		compilepattern(cs, pat->subpat2, &loc1, fail, canfail, istmp);
		break;
	case PatternParse:
		assert(isptrast(val->type));
		tl = pat->type;
		tr = pat->valuep->type_;
		// Evaluate exprs into successive tmp slots.
		memset(&loc1, 0, sizeof loc1);
		loc = &loc1;
		cs->pushtmp();
		cmp = cs->mktmp(typebool());
		for(int i=0; i<pat->valuep->exprs_.size(); i++){
			e = pat->valuep->exprs_[i];
			tmp = cs->mktmp(e->type_);
			if(loc == &loc1)
				loc = tmp;
			cs->pushtmp();
			cs->leaexpr(e, tmp);
			cs->poptmp();
		}
		cs->emitC(pat->line, xdestructure);
		cs->emitloc(cmp);
		cs->emitloc(val);
		if(!isverticalsubtype(tl->left, tr->left))
			cs->emitp(tr->left);  // yes, we need the dynamic check
		else
			cs->emitp(nil);
		emitpatternast(cs, pat->valuep);
		cs->emitloc(loc);
		cs->emitp(pat->valuep->canonicals_);
		// Fail if not matched
		cs->emitjmp(pat->line, xncjmp, cmp, fail);
		*canfail = 1;
		cs->poptmp();
		// Matched: set the variable (if any)
		if(pat->var)
			cs->emitmov(pat->line, var2loc(cs, pat->var), val);
		break;
	case PatternConst:
		cs->pushtmp();
		// Evaluate the constant
		// XXX replace with evaltmp?
		tmp = cs->mktmp(pat->type);
		cs->evalinto(pat->value, tmp);
		// Test for equality
		cmp = cs->mktmp(typeint());
		cs->emitCLLL(pat->line, xeqi, cmp, val, tmp);
		// Fail if not equal
		cs->emitjmp(pat->line, xncjmp, cmp, fail);
		*canfail = 1;
		cs->poptmp();
		break;
	case PatternLit:
		if(!pat->var)
			break;
		if(pat->isptr){
			if(istmp){
				// Hmm.  Generally, the use a pointer
				// pattern indicates the intent to
				// mutate the extracted value, but not
				// necessarily.
				error(pat->line, "taking pointer to temporary value in pattern");
			}
			cs->emitCLL(pat->line, xlea, var2loc(cs, pat->var), val);
		}else{
			cs->emitmov(pat->line, var2loc(cs, pat->var), val);
		}
		break;
	case PatternAny:
		// Always succeeds and binds nothing, so do nothing
		break;
	case PatternDisjunct:
		subfail = mklabel(cs->tc_);
		subsucceed = mklabel(cs->tc_);
		subcanfail = subcanfail2 = 0;
		compilepattern(cs, pat->subpat, val, subfail, &subcanfail, istmp);
		cs->emitjmp(pat->line, xjmp, nil, subsucceed);
		cs->emitlabel(subfail);
		if (subcanfail) {
			compilepattern(cs, pat->subpat2, val, fail,
				       &subcanfail2, istmp);
		} else {
			error(pat->line, "left pattern cannot fail");
		}
		cs->emitlabel(subsucceed);
		if (subcanfail && subcanfail2)
			*canfail = 1;
		break;
	}
}

