struct ExprName;
struct ExprDeclEq;
struct ExprParse;
struct ExprStar;
struct ExprArray;
struct ExprCall;
struct ExprDot;
struct ExprInitField;
struct ExprTuple;
struct Loc;

struct Expr
{
	Line line_;
	Type *type_;
	Expr *expanded_;

	Type *infertype_;  // turd
	Type *xtype_;  // another turd
	
	Expr() :
		type_(nil), expanded_(nil),
		infertype_(nil), xtype_(nil) 
	{
		line_.file = nil;
		line_.line = 0;
	}
	
	virtual ~Expr();	
	virtual void typecheck(Typecheck*, int);
	virtual Loc *compile(Compile*, Loc*, Loc*);
	virtual Loc *lea(Compile*, int);
	virtual void pprint(Fmt*) = 0;
	
	// clumsy rtti substitute
	virtual ExprName *exprname() { return nil; }
	virtual ExprDeclEq *exprdecleq() { return nil; }
	virtual ExprParse *exprparse() { return nil; }
	virtual ExprStar *exprstar() { return nil; }
	virtual ExprArray *exprarray() { return nil; }
	virtual ExprCall *exprcall() { return nil; }
	virtual ExprDot *exprdot() { return nil; }
	virtual ExprTuple *exprtuple() { return nil; }
	virtual ExprInitField *exprinitfield() { return nil; }
};

struct ExprName : public Expr
{
	ExprName(Name name) :
		Expr(), name_(name),
		var_(nil) {}
	
	Name name_;
	
	Var *var_;
	
	void typecheck(Typecheck*, int);
	ExprName *exprname() { return this; }
	Loc *compile(Compile*, Loc*, Loc*);
	Loc *lea(Compile*, int);
	void pprint(Fmt*);
};

struct ExprNew : public Expr
{
	ExprNew(Expr *left) : 
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprStar : public Expr
{
	ExprStar(Expr *left) : 
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	Loc *lea(Compile*, int);
	ExprStar *exprstar() { return this; }
	void pprint(Fmt*);
};

struct ExprAddr : public Expr
{
	ExprAddr(Expr *left) : 
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprString : public Expr
{
	ExprString(char *str, int len) : 
		Expr(), str_(str), len_(len) {}
	
	char *str_;
	int len_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprBool : public Expr
{
	ExprBool(bool b) :
		Expr(), b_(b) {}
	bool b_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprInteger : public Expr
{
	ExprInteger(int i) :
		Expr(), i_(i) {}
	int i_;

	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprFloat : public Expr
{
	ExprFloat(double f) :
		Expr(), f_(f) {}
	double f_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprEq : public Expr
{
	ExprEq(Expr *left, Expr *right) :
		Expr(), left_(left), right_(right) {}
	
	Expr *left_;
	Expr *right_;

	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprDeclEq : public Expr
{
	ExprDeclEq(Expr *left, Type *type, Expr *right) :
		Expr(), left_(left), right_(right),
		flags_(0) { type_ = type;; }
	
	Expr *left_;
	Expr *right_;
	
	int flags_;  // DeclareConst
	
	void typecheck(Typecheck*, int);
	ExprDeclEq *exprdecleq() { return this; }
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprFn : public Expr
{
	ExprFn(Fn *fn, List<Name> *typevars) :
		Expr(), fn_(fn), typevars_(typevars) {}
	
	Fn *fn_;
	List<Name> *typevars_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprDot : public Expr
{
	ExprDot(Expr *left, Name name) :
		Expr(), left_(left), name_(name),
		var_(nil) {}
	
	Expr *left_;
	Name name_;
	
	Var *var_;  // structure member
	
	void typecheck(Typecheck*, int);
	ExprDot *exprdot() { return this; }
	Loc *compile(Compile*, Loc*, Loc*);
	Loc *lea(Compile*, int);
	void pprint(Fmt*);
};

struct ExprInit : public Expr
{
	ExprInit(Expr *left, const vector<ExprInitField*>& exprs) :
		Expr(), left_(left), exprs_(exprs),
		callop_(CallRegular) {}
	
	Expr *left_;
	vector<ExprInitField*> exprs_;
	
	CallOp callop_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprInitField : public Expr
{
	ExprInitField(Name name, Expr *left) :
		Expr(), name_(name), left_(left),
		var_(nil), xtype_(nil) {}
	
	Name name_;
	Expr *left_;
	
	Var *var_;
	Type *xtype_;
	
	ExprInitField *exprinitfield() { return this; }
	void pprint(Fmt*);
};

struct ExprParseString : public Expr
{
	ExprParseString(List<Name> *names, const vector<Expr*>& exprs) :
		Expr(), names_(names), exprs_(exprs) {}
	
	List<Name> *names_;
	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprParse : public Expr
{
	ExprParse(List<Name> *names, List<PToken*> *ptokens) :
		Expr(), names_(names), ptokens_(ptokens), canonicals_(nil),
		ast_(nil) {}
	
	List<Name> *names_;
	List<PToken*> *ptokens_;
	List<Var*> *canonicals_;
	
	Ast *ast_;
	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	ExprParse *exprparse() { return this; }
	void pprint(Fmt*);
};

struct ExprDestructure : public Expr
{
	ExprDestructure(Expr *left, ExprParse *right) :
		left_(left), right_(right) {}
	
	Expr *left_;
	ExprParse *right_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprDestructType : public Expr
{
	ExprDestructType(Expr *left, List<Name> *names) :
		Expr(), left_(left), names_(names) {}
	
	Expr *left_;
	List<Name> *names_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprRestructure : public Expr
{
	ExprRestructure(ExprParse *left, Expr *right) :
		left_(left), right_(right) {}
	
	ExprParse *left_;
	Expr *right_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprSameStructure : public Expr
{
	ExprSameStructure(Expr *left, Expr *right) :
		left_(left), right_(right) {}
	
	Expr *left_;
	Expr *right_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprAstParent : public Expr
{
	ExprAstParent(Expr *left, List<Name> *names) :
		Expr(), left_(left), names_(names) {}

	Expr *left_;
	List<Name> *names_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprConvert : public Expr
{
	ExprConvert(Expr *left, Type *type) :
		Expr(), left_(left) { type_ = type;}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprNil : public Expr
{
	ExprNil() : Expr() {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprCall : public Expr
{
	ExprCall(Expr *left, const vector<Expr*>& exprs) :
		Expr(), left_(left), exprs_(exprs),
		callop_(CallRegular) {}
	
	Expr *left_;
	vector<Expr*> exprs_;
	
	CallOp callop_;
	
	void typecheck(Typecheck*, int);
	ExprCall *exprcall() { return this; }
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprTuple : public Expr
{
	ExprTuple(const vector<Expr*>& exprs) :
		Expr(), exprs_(exprs) {}

	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	ExprTuple *exprtuple() { return this; }
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprMkList : public Expr
{
	ExprMkList(const vector<Expr*>& exprs) :
		Expr(), exprs_(exprs) {}

	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprMkArray : public Expr
{
	ExprMkArray(const vector<Expr*>& exprs) :
		Expr(), exprs_(exprs) {}

	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprMkListOrArray : public Expr
{
	ExprMkListOrArray(const vector<Expr*>& exprs) :
		Expr(), exprs_(exprs) {}

	vector<Expr*> exprs_;
	
	void typecheck(Typecheck*, int);
	void pprint(Fmt*);
};

struct ExprMap : public Expr
{
	ExprMap(Expr *left, Expr *right) :
		Expr(), left_(left), right_(right) {}
	
	Expr *left_;
	Expr *right_;
	void pprint(Fmt*);
};

struct ExprMkMap : public Expr
{
	ExprMkMap(const vector<ExprMap*>& exprs) :
		Expr(), exprs_(exprs) {}

	vector<ExprMap*> exprs_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprListCons : public Expr
{
	ExprListCons(Expr *left, Expr *right) :
		Expr(), left_(left), right_(right) {}
	
	Expr *left_;
	Expr *right_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprLength : public Expr
{
	ExprLength(Expr *left) :
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprReverse : public Expr
{
	ExprReverse(Expr *left) :
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprTag : public Expr
{
	ExprTag(Expr *left) :
		Expr(), left_(left) {}
	
	Expr *left_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprSlice : public Expr
{
	ExprSlice(Expr *left, Expr *middle, Expr *right) :
		Expr(), left_(left), middle_(middle), right_(right) {}
	
	Expr *left_;
	Expr *middle_;
	Expr *right_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprBinary : public Expr
{
	ExprBinary(Expr *left, Expr *right) :
		Expr(), left_(left), right_(right) {}
	
	Expr *left_;
	Expr *right_;
};


struct ExprArray : public ExprBinary
{
	ExprArray(Expr *left, Expr *right) :
		ExprBinary(left, right) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	Loc *lea(Compile*, int);
	void pprint(Fmt*);
	ExprArray *exprarray() { return this; }
};

struct ExprAndAnd : public ExprBinary
{
	ExprAndAnd(Expr *left, Expr *right) :
		ExprBinary(left, right) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprOrOr : public ExprBinary
{
	ExprOrOr(Expr *left, Expr *right) :
		ExprBinary(left, right) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprEqEq : public ExprBinary
{
	ExprEqEq(Expr *left, Expr *right) :
		ExprBinary(left, right) {}

	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprNotEq : public ExprBinary
{
	ExprNotEq(Expr *left, Expr *right) :
		ExprBinary(left, right) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};

struct ExprIncDec : public Expr
{
	ExprIncDec(Expr *left, int post, int inc) :
		Expr(), left_(left), post_(post), inc_(inc) {}
	
	Expr *left_;
	int post_;
	int inc_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);
};


// Table-based handling of simple op-codes.
// Eventually this will go away in favor of a better type system
// and a more general Expr type.

struct ExprSig
{
	Code *code;
	Type *left;
	Type *right;
	Type *out;
};

struct ExprPreInc : public ExprIncDec
{
	ExprPreInc(Expr *left) :
		ExprIncDec(left, 0, 1) {}
};

struct ExprPreDec : public ExprIncDec
{
	ExprPreDec(Expr *left) :
		ExprIncDec(left, 0, -1) {}
};

struct ExprPostInc : public ExprIncDec
{
	ExprPostInc(Expr *left) :
		ExprIncDec(left, 1, 1) {}
};

struct ExprPostDec : public ExprIncDec
{
	ExprPostDec(Expr *left) :
		ExprIncDec(left, 1, -1) {}
};

struct ExprCode : public ExprBinary
{
	ExprCode(Expr *left, Expr *right, int assign) :
		ExprBinary(left, right), assign_(assign),
		code_(nil) {}
	
	int assign_;
	
	Code *code_;
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	void pprint(Fmt*);

	virtual const vector<ExprSig>& sigs();
	virtual const char *oper() = 0;
};

struct ExprAdd : public ExprCode
{
	ExprAdd(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	const vector<ExprSig>& sigs();
	const char *oper() { return "+"; }
};

struct ExprAddEq : public ExprCode
{
	ExprAddEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	void typecheck(Typecheck*, int);
	Loc *compile(Compile*, Loc*, Loc*);
	const vector<ExprSig>& sigs();
	const char *oper() { return "+="; }
};

struct ExprSub : public ExprCode
{
	ExprSub(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "-"; }
};

struct ExprSubEq : public ExprCode
{
	ExprSubEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "-="; }
};

struct ExprMul : public ExprCode
{
	ExprMul(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "*"; }
};

struct ExprMulEq : public ExprCode
{
	ExprMulEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "*="; }
};

struct ExprDiv : public ExprCode
{
	ExprDiv(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "/"; }
};

struct ExprDivEq : public ExprCode
{
	ExprDivEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "/="; }
};

struct ExprMod : public ExprCode
{
	ExprMod(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "%"; }
};

struct ExprModEq : public ExprCode
{
	ExprModEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "%="; }
};

struct ExprAnd : public ExprCode
{
	ExprAnd(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "&"; }
};

struct ExprAndEq : public ExprCode
{
	ExprAndEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "&="; }
};

struct ExprOr : public ExprCode
{
	ExprOr(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "|"; }
};

struct ExprOrEq : public ExprCode
{
	ExprOrEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "|="; }
};

struct ExprXor : public ExprCode
{
	ExprXor(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "^"; }
};

struct ExprXorEq : public ExprCode
{
	ExprXorEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "^="; }
};

struct ExprLsh : public ExprCode
{
	ExprLsh(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "<<"; }
};

struct ExprLshEq : public ExprCode
{
	ExprLshEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "<<="; }
};

struct ExprRsh : public ExprCode
{
	ExprRsh(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return ">>"; }
};

struct ExprRshEq : public ExprCode
{
	ExprRshEq(Expr *left, Expr *right) :
		ExprCode(left, right, 1) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return ">>="; }
};

struct ExprLt : public ExprCode
{
	ExprLt(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "<"; }
};

struct ExprLe : public ExprCode
{
	ExprLe(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "<="; }
};

struct ExprGt : public ExprCode
{
	ExprGt(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return ">"; }
};

struct ExprGe : public ExprCode
{
	ExprGe(Expr *left, Expr *right) :
		ExprCode(left, right, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return ">="; }
};

struct ExprTwid : public ExprCode
{
	ExprTwid(Expr *left) :
		ExprCode(left, nil, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "~"; }
};

struct ExprPos : public ExprCode
{
	ExprPos(Expr *left) :
		ExprCode(left, nil, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "+"; }
};

struct ExprNeg : public ExprCode
{
	ExprNeg(Expr *left) :
		ExprCode(left, nil, 0) {}
	
	const vector<ExprSig>& sigs();
	const char *oper() { return "-"; }
};

struct ExprNot : public ExprCode
{
	ExprNot(Expr *left) :
		ExprCode(left, nil, 0) {}
	
	void typecheck(Typecheck*, int);
	const vector<ExprSig>& sigs();
	const char *oper() { return "!"; }
};

