#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "list2.h"
#include "cfg.h"
#include "emalloc.h"

/*
 * Bit sets
 */
Xset*
xxnewset(uint n)
{
	Xset *s;
	int nw;

	nw = (n+YW-1)/YW;
	s = emalloc(sizeof(Xset)+(nw-1)*sizeof(uint));
	s->n = nw;
	return s;
}

void
xxfreeset(Xset *s)
{
	efree(s);
}

static int
bits(uint b)
{
	uint n;

	n = b;
	n = (n&0x55555555)+((n&0xAAAAAAAA)>>1);
	n = (n&0x33333333)+((n&0xCCCCCCCC)>>2);
	n = (n&0x0F0F0F0F)+((n&0xF0F0F0F0)>>4);
	n = (n&0x00FF00FF)+((n&0xFF00FF00)>>8);
	n = (n&0x0000FFFF)+((n&0xFFFF0000)>>16);
	return n;
}

int
xxcountset(Xset *s)
{
	int i, n;

	assert(s != nil);
	n = 0;
	for(i=0; i<s->n; i++)
		n += bits(s->a[i]);
	return n;
}

int
xxaddset(Xset *s, Xset *t)
{
	int i, did;

	assert(s != nil);
	assert(t != nil);
	assert(s->n == t->n);
	did = 0;
	for(i=0; i<s->n; i++){
		if(~s->a[i] & t->a[i])
			did = 1;
		s->a[i] |= t->a[i];
	}
	return did;
}

int
xxinset(Xset *s, int x)
{
	assert(s != nil);
	return s->a[x/YW] & (1<<(x%YW));
}

int
xxaddbit(Xset *s, int x)
{
	int b;
	
	assert(s != nil);
	assert(0 <= x && x < s->n*YW);
	b = 1<<(x%YW);
	if(s->a[x/YW] & b)
		return 1;
	s->a[x/YW] |= b;
	return 0;
}

int
xxnextset(Xset *s, int i)
{
	if(i>=0)
		for(; i<s->n*YW; i++)
			if(s->a[i/YW]&(1<<(i%YW)))
				return i;
	return -1;
}
