#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "arena.h"
#include "emalloc.h"
#include "list2.h"
#include "cfg.h"

static uint
hash(char *str)
{
	/* http://www.cs.yorku.ca/~oz/hash.html */
	uint32 h;
	int c;

	h = 5381;
	for(; (c=(uchar)*str) != 0; str++)
		h = ((h << 5) + h) ^ c; /* hash * 33 ^ c */
	return h;
}

CfgSym<void>*
_cfgnewsym(CfgGram<void> *g, char *name, char *type)
{
	int i, sz;
	CfgSym<void> *s;

	if(name){
		for(i=0; i<g->nsym; i++){
			s = g->sym[i];
			if(s->name && strcmp(s->name, name) == 0){
				if((type==nil || s->type==nil) && (type!=nil || s->type!=nil))
					goto bad;
				return s;
			}
		}
	}

	s = (CfgSym<void>*)emalloc(sizeof *s);
	s->gram = g;
	s->n = g->nsym;
	if(name){
		s->name = nameof(name);
		s->code = hash(name);
	}
	if(type)
		s->type = nameof(type);

	if(g->nsym%32 == 0){
		sz = sizeof g->sym[0];
		g->sym = (CfgSym<void>**)erealloczero(g->sym, g->nsym*sz, (g->nsym+32)*sz);
	}
	g->sym[g->nsym++] = s;
	return s;

bad:
//	werrstr("bad redeclaration of symbol %s", name);
//	if(g->debug)
//XXX		fprint(2, "cfgnewsym: %r\n");
	return nil;
}

CfgSym<void>*
_cfglooksym(CfgGram<void> *g, char *name)
{
	int i;

	if(name == nil)
		return nil;

	name = nameof(name);
	for(i=0; i<g->nsym; i++)
		if(g->sym[i]->name == name)
			return g->sym[i];
//	werrstr("unknown symbol %s", name);
//	if(g->debug)
//		fprint(2, "%r\n");
	return nil;
}

