struct Arena;

template<typename T> 
struct List
{
	T x;
	List<T> *next;
	List<T> *end;
	Arena *arena;
};

template<typename T> List<T> *list1(struct Arena*, T);
template<typename T> List<T> *list2(struct Arena*, T, T);
template<typename T> List<T> *list3(struct Arena*, T, T, T);
template<typename T> List<T> *list4(struct Arena*, T, T, T, T);
template<typename T> List<T> *list5(struct Arena*, T, T, T, T, T);
template<typename T> List<T> *listn(struct Arena*, int, ...);
template<typename T> List<T> *listappend(struct Arena*, List<T>*, T);
template<typename T> void listfree(List<T>*);
template<typename T> List<T> *listinsert(struct Arena*, List<T>*, T);
template<typename T> List<T> *listcopy(struct Arena*, List<T>*);
template<typename T> List<T> *listsplit(List<T>*, List<T>*);
template<typename T> List<T> *listconcat(List<T> *l, List<T> *after);
template<typename T> List<T> *listcopyreverse(struct Arena*, List<T>*);
template<typename T> int listeq(List<T>*, List<T>*);
template<typename T> int listlen(List<T>*);
template<typename T> T listpeek(List<T>*);
template<typename T> List<T> *listpop(List<T>*);
template<typename T> List<T> *listsort(List<T>*, int (*cmp)(T,T));
template<typename T> int listfind(List<T>*, const T);
template<typename T> T listnth(List<T>*, int);
template<typename T> List<T> *listmap(struct Arena*, List<T>*, T (*mapfn)(T));
template<typename T> List<T> *listremoveif(List<T>*, List<T>**, int (*removeif)(T));
template<typename T> List<T> *listremovedups(List<T>*, List<T>**, int (*cmp)(T,T));

#define ZList (nil)

template<typename T>
List<T>*
list1(struct Arena *a, T x)
{
	List<T> *l;
	if(a)
		l = (List<T>*)amalloc(a, sizeof *l);
	else
		l = (List<T>*)emalloc(sizeof *l);
	l->arena = a;
	l->x = x;
	l->next = nil;
	l->end = l;
	return l;
}

template<typename T>
List<T>*
list2(Arena *a, T x, T y)
{
	return listappend(a, list1(a, x), y);
}

template<typename T>
List<T>*
list3(Arena *a, T x, T y, T z)
{
	return listappend(a, list2(a, x, y), z);
}

template<typename T>
List<T>*
list4(Arena *a, T x, T y, T z, T w)
{
	return listappend(a, list3(a, x, y, z), w);
}

template<typename T>
List<T>*
list5(Arena *a, T x, T y, T z, T w, T q)
{
	return listappend(a, list4(a, x, y, z, w), q);
}

template<typename T>
List<T>*
listn(Arena* a, int n, ...)
{
	va_list va;
	List<T> *l;

	va_start(va, n);
	l = ZList;
	while(n-- > 0)
		l = listappend(a, l, va_arg(va, T));
	va_end(va);
	return l;
}

template<typename T>
List<T>*
listappend(Arena *a, List<T> *l, T x)
{
	if(l == nil)
		return list1(a, x);
	
	l->end->next = list1(a, x);
	l->end = l->end->next;
	return l;
}

template<typename T>
List<T>*
listinsert(Arena *a, List<T> *l, T x)
{
	List<T> *l1;
	
	l1 = list1(a, x);
	if(l){
		l1->next = l;
		l1->end = l->end;
	}
	return l1;
}

template<typename T>
void
listfree(List<T> *l)
{
	List<T> *next;
	
	for(; l; l=next){
		next = l->next;
		if(l->arena == nil)
			free(l);
	}
}

template<typename T>
List<T>*
listcopy(Arena *a, List<T> *l)
{
	List<T> *ll;
	
	ll = ZList;
	for(; l; l=l->next)
		ll = listappend(a, ll, l->x);
	return ll;
}

template<typename T>
List<T>*
listsplit(List<T> *l, List<T> *end)
{
	List<T> *ll;

	if (l == nil)
		return nil;
	if (l == end)
		return ZList;

	for (ll=l; ll->next!=end; ll=ll->next);
	ll->next = ZList;
	l->end = ll;
	return l;
}

template<typename T>
List<T>*
listcopyreverse(Arena *a, List<T> *l)
{
	List<T> *ll;
	
	ll = ZList;
	for(; l; l=l->next)
		ll = listinsert(a, ll, l->x);
	return ll;
}

template<typename T>
List<T>*
listconcat(List<T> *l, List<T> *after)
{
	if(l == nil)
		return after;
	if(after == nil)
		return l;
	
	l->end->next = after;
	l->end = after->end;
	return l;
}

template<typename T>
int
listeq(List<T> *a, List<T> *b)
{
	if(a == b)
		return 1;
	for(; a && b; a=a->next, b=b->next)
		if(a->x != b->x)
			return 0;
	if(a || b)
		return 0;
	return 1;
}

template<typename T>
int
listlen(List<T> *l)
{
	int n;
	
	n = 0;
	for(; l; l=l->next)
		n++;
	return n;
}

template<typename T>
List<T>*
listpop(List<T> *l)
{
	List<T> *ll;

	if(l == nil)
		return nil;
	ll = l->next;
	if(l->arena == nil)
		free(l);
	return ll;
}

template<typename T>
T
listpeek(List<T> *l)
{
	if(l == nil)
		return nil;
	return l->x;
}

template<typename T>
int
listfind(List<T> *l, const T x)
{
	for(; l; l=l->next)
		if(l->x == x)
			return 1;
	return 0;
}

template<typename T>
T
listnth(List<T> *l, int n)
{
	if(n < 0)
		return nil;
	for(; n && l; n--, l=l->next)
		;
	return l ? l->x : nil;
}

template<typename T>
List<T>*
_listmerge(List<T> *l1, List<T> *l2, int (*cmp)(const T,const T))
{
	int i;
	List<T> lhead, *l;

	l = &lhead;
	while(l1 || l2){
		if(l1 == nil)
			i = 1;
		else if(l2 == nil)
			i = -1;
		else{
			if(cmp)
				i = (*cmp)(l1->x, l2->x);
			else if(l1->x <= l2->x)
				i = -1;
			else
				i = 1;
		}
		if(i <= 0){
			l->next = l1;
			l1 = l1->next;
		}else{
			l->next = l2;
			l2 = l2->next;
		}
		l = l->next;
	}
	l->next = nil;
	return lhead.next;
}

template<typename T>
List<T>*
listsort(List<T> *l, int (*cmp)(const T,const T))
{
	List<T> *l1, *l2;
	List<T> *l1head, *l2head;

	if(l == nil)
		return nil;
	if(l->next == nil)
		return l;

	l1head = l1 = l;
	l2head = l2 = l->next;
	l = l2->next;
	while(l){
		l1->next = l;
		l1 = l;
		if((l = l->next) != nil){
			l2->next = l;
			l2 = l;
			l = l->next;
		}
	}
	l1->next = nil;
	l2->next = nil;

	l1 = listsort(l1head, cmp);
	l2 = listsort(l2head, cmp);
	return _listmerge(l1, l2, cmp);
}

template<typename T, typename V>
List<V>*
listmap(Arena *a, List<T> *l, V (*f)(const T))
{
	List<V> *ll;
	
	assert(f);
	ll = ZList;
	for(; l; l=l->next)
		ll = listappend(a, ll, f(l->x));
	return ll;
}

template<typename T>
List<T>*
listfilter(List<T> *l, int (*keep)(const T))
{
	assert(keep);
	
	List<T> *out = nil;
	for(; l; l=l->next){
		if(keep(l->x))
			out = listappend((Arena*)nil, out, l->x);
	}
	return out;
}

template<typename T>
List<T>*
listremovedups(List<T> *l, List<T> **pdups, int (*cmp)(const T,const T))
{
	List<T> *ll, *prev, *dups;
	int i;

	if(l == nil || l->next == nil){
		if(pdups)
			*pdups = nil;
		return l;
	}
	ll = l->next;
	prev = l;
	dups = nil;

	// ASSUMES THE LIST IS SORTED!
	// Remove duplicate elements from the list and place them in 'dups'.
	for(; ll; ll=ll->next){
		if(cmp)
			i = (*cmp)(prev->x, ll->x);
		else
			i = (prev->x == ll->x) ? 0 : 1;
		if(i == 0){
			prev->next = ll->next;
			ll->next = nil;
			ll->end = ll;
			dups = listconcat(dups, ll);
		}
		else
			prev = ll;
	}

	l->end = prev;
	if(pdups)
		*pdups = dups;
	else
		listfree(dups);
	return l;
}

